
#ifndef TTKMERGETREE_EXPORT_H
#define TTKMERGETREE_EXPORT_H

#ifdef TTKMERGETREE_STATIC_DEFINE
#  define TTKMERGETREE_EXPORT
#  define TTKMERGETREE_NO_EXPORT
#else
#  ifndef TTKMERGETREE_EXPORT
#    ifdef ttkMergeTree_EXPORTS
        /* We are building this library */
#      define TTKMERGETREE_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKMERGETREE_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKMERGETREE_NO_EXPORT
#    define TTKMERGETREE_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKMERGETREE_DEPRECATED
#  define TTKMERGETREE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMERGETREE_DEPRECATED_EXPORT
#  define TTKMERGETREE_DEPRECATED_EXPORT TTKMERGETREE_EXPORT TTKMERGETREE_DEPRECATED
#endif

#ifndef TTKMERGETREE_DEPRECATED_NO_EXPORT
#  define TTKMERGETREE_DEPRECATED_NO_EXPORT TTKMERGETREE_NO_EXPORT TTKMERGETREE_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMERGETREE_NO_DEPRECATED
#    define TTKMERGETREE_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkMergeTreeModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKMERGETREE_EXPORT_H */
