
#ifndef TTKTRACKINGFROMOVERLAP_EXPORT_H
#define TTKTRACKINGFROMOVERLAP_EXPORT_H

#ifdef TTKTRACKINGFROMOVERLAP_STATIC_DEFINE
#  define TTKTRACKINGFROMOVERLAP_EXPORT
#  define TTKTRACKINGFROMOVERLAP_NO_EXPORT
#else
#  ifndef TTKTRACKINGFROMOVERLAP_EXPORT
#    ifdef ttkTrackingFromOverlap_EXPORTS
        /* We are building this library */
#      define TTKTRACKINGFROMOVERLAP_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKTRACKINGFROMOVERLAP_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKTRACKINGFROMOVERLAP_NO_EXPORT
#    define TTKTRACKINGFROMOVERLAP_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKTRACKINGFROMOVERLAP_DEPRECATED
#  define TTKTRACKINGFROMOVERLAP_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTRACKINGFROMOVERLAP_DEPRECATED_EXPORT
#  define TTKTRACKINGFROMOVERLAP_DEPRECATED_EXPORT TTKTRACKINGFROMOVERLAP_EXPORT TTKTRACKINGFROMOVERLAP_DEPRECATED
#endif

#ifndef TTKTRACKINGFROMOVERLAP_DEPRECATED_NO_EXPORT
#  define TTKTRACKINGFROMOVERLAP_DEPRECATED_NO_EXPORT TTKTRACKINGFROMOVERLAP_NO_EXPORT TTKTRACKINGFROMOVERLAP_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTRACKINGFROMOVERLAP_NO_DEPRECATED
#    define TTKTRACKINGFROMOVERLAP_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkTrackingFromOverlapModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKTRACKINGFROMOVERLAP_EXPORT_H */
