
#ifndef TTKPERSISTENTGENERATORS_EXPORT_H
#define TTKPERSISTENTGENERATORS_EXPORT_H

#ifdef TTKPERSISTENTGENERATORS_STATIC_DEFINE
#  define TTKPERSISTENTGENERATORS_EXPORT
#  define TTKPERSISTENTGENERATORS_NO_EXPORT
#else
#  ifndef TTKPERSISTENTGENERATORS_EXPORT
#    ifdef ttkPersistentGenerators_EXPORTS
        /* We are building this library */
#      define TTKPERSISTENTGENERATORS_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKPERSISTENTGENERATORS_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKPERSISTENTGENERATORS_NO_EXPORT
#    define TTKPERSISTENTGENERATORS_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKPERSISTENTGENERATORS_DEPRECATED
#  define TTKPERSISTENTGENERATORS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPERSISTENTGENERATORS_DEPRECATED_EXPORT
#  define TTKPERSISTENTGENERATORS_DEPRECATED_EXPORT TTKPERSISTENTGENERATORS_EXPORT TTKPERSISTENTGENERATORS_DEPRECATED
#endif

#ifndef TTKPERSISTENTGENERATORS_DEPRECATED_NO_EXPORT
#  define TTKPERSISTENTGENERATORS_DEPRECATED_NO_EXPORT TTKPERSISTENTGENERATORS_NO_EXPORT TTKPERSISTENTGENERATORS_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPERSISTENTGENERATORS_NO_DEPRECATED
#    define TTKPERSISTENTGENERATORS_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkPersistentGeneratorsModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKPERSISTENTGENERATORS_EXPORT_H */
