
#ifndef TTKMERGEBLOCKTABLES_EXPORT_H
#define TTKMERGEBLOCKTABLES_EXPORT_H

#ifdef TTKMERGEBLOCKTABLES_STATIC_DEFINE
#  define TTKMERGEBLOCKTABLES_EXPORT
#  define TTKMERGEBLOCKTABLES_NO_EXPORT
#else
#  ifndef TTKMERGEBLOCKTABLES_EXPORT
#    ifdef ttkMergeBlockTables_EXPORTS
        /* We are building this library */
#      define TTKMERGEBLOCKTABLES_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKMERGEBLOCKTABLES_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKMERGEBLOCKTABLES_NO_EXPORT
#    define TTKMERGEBLOCKTABLES_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKMERGEBLOCKTABLES_DEPRECATED
#  define TTKMERGEBLOCKTABLES_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMERGEBLOCKTABLES_DEPRECATED_EXPORT
#  define TTKMERGEBLOCKTABLES_DEPRECATED_EXPORT TTKMERGEBLOCKTABLES_EXPORT TTKMERGEBLOCKTABLES_DEPRECATED
#endif

#ifndef TTKMERGEBLOCKTABLES_DEPRECATED_NO_EXPORT
#  define TTKMERGEBLOCKTABLES_DEPRECATED_NO_EXPORT TTKMERGEBLOCKTABLES_NO_EXPORT TTKMERGEBLOCKTABLES_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMERGEBLOCKTABLES_NO_DEPRECATED
#    define TTKMERGEBLOCKTABLES_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkMergeBlockTablesModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKMERGEBLOCKTABLES_EXPORT_H */
