
#ifndef TTKREEBGRAPH_EXPORT_H
#define TTKREEBGRAPH_EXPORT_H

#ifdef TTKREEBGRAPH_STATIC_DEFINE
#  define TTKREEBGRAPH_EXPORT
#  define TTKREEBGRAPH_NO_EXPORT
#else
#  ifndef TTKREEBGRAPH_EXPORT
#    ifdef ttkReebGraph_EXPORTS
        /* We are building this library */
#      define TTKREEBGRAPH_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKREEBGRAPH_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKREEBGRAPH_NO_EXPORT
#    define TTKREEBGRAPH_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKREEBGRAPH_DEPRECATED
#  define TTKREEBGRAPH_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKREEBGRAPH_DEPRECATED_EXPORT
#  define TTKREEBGRAPH_DEPRECATED_EXPORT TTKREEBGRAPH_EXPORT TTKREEBGRAPH_DEPRECATED
#endif

#ifndef TTKREEBGRAPH_DEPRECATED_NO_EXPORT
#  define TTKREEBGRAPH_DEPRECATED_NO_EXPORT TTKREEBGRAPH_NO_EXPORT TTKREEBGRAPH_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKREEBGRAPH_NO_DEPRECATED
#    define TTKREEBGRAPH_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkReebGraphModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKREEBGRAPH_EXPORT_H */
