
#ifndef TTKENDFOR_EXPORT_H
#define TTKENDFOR_EXPORT_H

#ifdef TTKENDFOR_STATIC_DEFINE
#  define TTKENDFOR_EXPORT
#  define TTKENDFOR_NO_EXPORT
#else
#  ifndef TTKENDFOR_EXPORT
#    ifdef ttkEndFor_EXPORTS
        /* We are building this library */
#      define TTKENDFOR_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKENDFOR_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKENDFOR_NO_EXPORT
#    define TTKENDFOR_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKENDFOR_DEPRECATED
#  define TTKENDFOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKENDFOR_DEPRECATED_EXPORT
#  define TTKENDFOR_DEPRECATED_EXPORT TTKENDFOR_EXPORT TTKENDFOR_DEPRECATED
#endif

#ifndef TTKENDFOR_DEPRECATED_NO_EXPORT
#  define TTKENDFOR_DEPRECATED_NO_EXPORT TTKENDFOR_NO_EXPORT TTKENDFOR_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKENDFOR_NO_DEPRECATED
#    define TTKENDFOR_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkEndForModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKENDFOR_EXPORT_H */
