
#ifndef TTKFLATTENMULTIBLOCK_EXPORT_H
#define TTKFLATTENMULTIBLOCK_EXPORT_H

#ifdef TTKFLATTENMULTIBLOCK_STATIC_DEFINE
#  define TTKFLATTENMULTIBLOCK_EXPORT
#  define TTKFLATTENMULTIBLOCK_NO_EXPORT
#else
#  ifndef TTKFLATTENMULTIBLOCK_EXPORT
#    ifdef ttkFlattenMultiBlock_EXPORTS
        /* We are building this library */
#      define TTKFLATTENMULTIBLOCK_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TTKFLATTENMULTIBLOCK_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TTKFLATTENMULTIBLOCK_NO_EXPORT
#    define TTKFLATTENMULTIBLOCK_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TTKFLATTENMULTIBLOCK_DEPRECATED
#  define TTKFLATTENMULTIBLOCK_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKFLATTENMULTIBLOCK_DEPRECATED_EXPORT
#  define TTKFLATTENMULTIBLOCK_DEPRECATED_EXPORT TTKFLATTENMULTIBLOCK_EXPORT TTKFLATTENMULTIBLOCK_DEPRECATED
#endif

#ifndef TTKFLATTENMULTIBLOCK_DEPRECATED_NO_EXPORT
#  define TTKFLATTENMULTIBLOCK_DEPRECATED_NO_EXPORT TTKFLATTENMULTIBLOCK_NO_EXPORT TTKFLATTENMULTIBLOCK_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKFLATTENMULTIBLOCK_NO_DEPRECATED
#    define TTKFLATTENMULTIBLOCK_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: ttkFlattenMultiBlockModule.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"

#endif /* TTKFLATTENMULTIBLOCK_EXPORT_H */
